@@ -26,7 +26,7 @@
  *           Sydney, 2052
  *           Australia
  *
- *    Additions for bitmap and write-behind RAID options, Copyright (C) 2003-2004, 
+ *    Additions for bitmap and write-behind RAID options, Copyright (C) 2003-2004,
  *    Paul Clements, SteelEye Technology, Inc.
  */
 
@@ -74,7 +74,9 @@ int main(int argc, char *argv[])
 	int brief = 0;
 	int force = 0;
 	int test = 0;
+	int export = 0;
 	int assume_clean = 0;
+	char *symlinks = NULL;
 	/* autof indicates whether and how to create device node.
 	 * bottom 3 bits are style.  Rest (when shifted) are number of parts
 	 * 0  - unset
@@ -100,10 +102,12 @@ int main(int argc, char *argv[])
 	int re_add = 0;
 	char *shortopt = short_options;
 	int dosyslog = 0;
+	int rebuild_map = 0;
 	int auto_update_home = 0;
 
 	int copies;
 	int print_help = 0;
+	FILE *outf;
 
 	int mdfd = -1;
 
@@ -129,7 +133,7 @@ int main(int argc, char *argv[])
 		/* firstly, some mode-independant options */
 		switch(opt) {
 		case 'h':
-			if (option_index > 0 && 
+			if (option_index > 0 &&
 			    strcmp(long_options[option_index].name, "help-options")==0)
 				print_help = 2;
 			else
@@ -156,6 +160,9 @@ int main(int argc, char *argv[])
 			}
 			continue;
 
+		case 'Y': export++;
+			continue;
+
 		case HomeHost:
 			homehost = optarg;
 			continue;
@@ -167,13 +174,13 @@ int main(int argc, char *argv[])
 		}
 		/* second, figure out the mode.
 		 * Some options force the mode.  Others
-		 * set the mode if it isn't already 
+		 * set the mode if it isn't already
 		 */
 
 		switch(opt) {
 		case '@': /* just incase they say --manage */
 			newmode = MANAGE;
-			shortopt = short_bitmap_auto_options;
+			shortopt = short_bitmap_options;
 			break;
 		case 'a':
 		case 'r':
@@ -181,7 +188,7 @@ int main(int argc, char *argv[])
 		case ReAdd: /* re-add */
 			if (!mode) {
 				newmode = MANAGE;
-				shortopt = short_bitmap_auto_options;
+				shortopt = short_bitmap_options;
 			}
 			break;
 
@@ -189,7 +196,13 @@ int main(int argc, char *argv[])
 		case 'B': newmode = BUILD; shortopt = short_bitmap_auto_options; break;
 		case 'C': newmode = CREATE; shortopt = short_bitmap_auto_options; break;
 		case 'F': newmode = MONITOR;break;
-		case 'G': newmode = GROW; shortopt = short_bitmap_auto_options; break;
+		case 'G': newmode = GROW;
+			shortopt = short_bitmap_options;
+			break;
+		case 'I': newmode = INCREMENTAL;
+			shortopt = short_bitmap_auto_options; break;
+		case AutoDetect:
+			newmode = AUTODETECT; break;
 
 		case '#':
 		case 'D':
@@ -200,6 +213,7 @@ int main(int argc, char *argv[])
 		case 'S':
 		case 'o':
 		case 'w':
+		case 'W':
 		case 'K': if (!mode) newmode = MISC; break;
 		}
 		if (mode && newmode == mode) {
@@ -219,10 +233,10 @@ int main(int argc, char *argv[])
 			mode = newmode;
 		} else {
 			/* special case of -c --help */
-			if (opt == 'c' && 
+			if (opt == 'c' &&
 			    ( strncmp(optarg, "--h", 3)==0 ||
 			      strncmp(optarg, "-h", 2)==0)) {
-				fputs(Help_config, stderr);
+				fputs(Help_config, stdout);
 				exit(0);
 			}
 
@@ -242,7 +256,7 @@ int main(int argc, char *argv[])
 					dv->next = NULL;
 					*devlistend = dv;
 					devlistend = &dv->next;
-			
+
 					devs_found++;
 					continue;
 				}
@@ -267,6 +281,8 @@ int main(int argc, char *argv[])
 		case 'C':
 		case 'F':
 		case 'G':
+		case 'I':
+		case AutoDetect:
 			continue;
 		}
 		if (opt == 1) {
@@ -292,7 +308,7 @@ int main(int argc, char *argv[])
 			dv->next = NULL;
 			*devlistend = dv;
 			devlistend = &dv->next;
-			
+
 			devs_found++;
 			continue;
 		}
@@ -319,6 +335,7 @@ int main(int argc, char *argv[])
 		case O(ASSEMBLE,AutoHomeHost):
 			auto_update_home = 1;
 			continue;
+		case O(INCREMENTAL, 'e'):
 		case O(CREATE,'e'):
 		case O(ASSEMBLE,'e'):
 		case O(MISC,'e'): /* set metadata (superblock) information */
@@ -326,7 +343,7 @@ int main(int argc, char *argv[])
 				fprintf(stderr, Name ": metadata information already given\n");
 				exit(2);
 			}
-			for(i=0; !ss && superlist[i]; i++) 
+			for(i=0; !ss && superlist[i]; i++)
 				ss = superlist[i]->match_metadata_desc(optarg);
 
 			if (!ss) {
@@ -435,7 +452,7 @@ int main(int argc, char *argv[])
 			case -5: /* Faulty
 				  * modeNNN
 				  */
-				    
+
 			{
 				int ln = strcspn(optarg, "0123456789");
 				char *m = strdup(optarg);
@@ -495,10 +512,17 @@ int main(int argc, char *argv[])
 
 		case O(CREATE,'a'):
 		case O(BUILD,'a'):
+		case O(INCREMENTAL,'a'):
 		case O(ASSEMBLE,'a'): /* auto-creation of device node */
 			autof = parse_auto(optarg, "--auto flag", 0);
 			continue;
 
+		case O(CREATE,Symlinks):
+		case O(BUILD,Symlinks):
+		case O(ASSEMBLE,Symlinks): /* auto creation of symlinks in /dev to /dev/md */
+			symlinks = optarg;
+			continue;
+
 		case O(BUILD,'f'): /* force honouring '-n 1' */
 		case O(GROW,'f'): /* ditto */
 		case O(CREATE,'f'): /* force honouring of device list */
@@ -508,6 +532,7 @@ int main(int argc, char *argv[])
 			continue;
 
 			/* now for the Assemble options */
+		case O(CREATE,'u'): /* uuid of array */
 		case O(ASSEMBLE,'u'): /* uuid of array */
 			if (ident.uuid_set) {
 				fprintf(stderr, Name ": uuid cannot be set twice.  "
@@ -561,7 +586,7 @@ int main(int argc, char *argv[])
 				exit(2);
 			}
 			update = optarg;
-			if (strcmp(update, "sparc2.2")==0) 
+			if (strcmp(update, "sparc2.2")==0)
 				continue;
 			if (strcmp(update, "super-minor") == 0)
 				continue;
@@ -575,12 +600,14 @@ int main(int argc, char *argv[])
 				continue;
 			if (strcmp(update, "homehost")==0)
 				continue;
+			if (strcmp(update, "devicesize")==0)
+				continue;
 			if (strcmp(update, "byteorder")==0) {
 				if (ss) {
 					fprintf(stderr, Name ": must not set metadata type with --update=byteorder.\n");
 					exit(2);
 				}
-				for(i=0; !ss && superlist[i]; i++) 
+				for(i=0; !ss && superlist[i]; i++)
 					ss = superlist[i]->match_metadata_desc("0.swap");
 				if (!ss) {
 					fprintf(stderr, Name ": INTERNAL ERROR cannot find 0.swap\n");
@@ -589,11 +616,24 @@ int main(int argc, char *argv[])
 
 				continue;
 			}
-			fprintf(stderr, Name ": '--update %s' invalid.  Only 'sparc2.2', 'super-minor', 'uuid', 'resync' or 'summaries' supported\n",update);
-			exit(2);
+			if (strcmp(update,"?") == 0 ||
+			    strcmp(update, "help") == 0) {
+				outf = stdout;
+				fprintf(outf, Name ": ");
+			} else {
+				outf = stderr;
+				fprintf(outf,
+					Name ": '--update=%s' is invalid.  ",
+					update);
+			}
+			fprintf(outf, "Valid --update options are:\n"
+		"     'sparc2.2', 'super-minor', 'uuid', 'name', 'resync',\n"
+		"     'summaries', 'homehost', 'byteorder', 'devicesize'.\n");
+			exit(outf == stdout ? 0 : 2);
 
 		case O(ASSEMBLE,NoDegraded): /* --no-degraded */
-			runstop = -1; /* --stop isn't allowed for --assemble, so we overload slightly */
+			runstop = -1; /* --stop isn't allowed for --assemble,
+				       * so we overload slightly */
 			continue;
 
 		case O(ASSEMBLE,'c'): /* config file */
@@ -611,6 +651,7 @@ int main(int argc, char *argv[])
 		case O(ASSEMBLE,'s'): /* scan */
 		case O(MISC,'s'):
 		case O(MONITOR,'s'):
+		case O(INCREMENTAL,'s'):
 			scan = 1;
 			continue;
 
@@ -663,7 +704,7 @@ int main(int argc, char *argv[])
 			test = 1;
 			continue;
 		case O(MONITOR,'y'): /* log messages to syslog */
-			openlog("mdadm", 0, SYSLOG_FACILITY);
+			openlog("mdadm", LOG_PID, SYSLOG_FACILITY);
 			dosyslog = 1;
 			continue;
 
@@ -685,6 +726,7 @@ int main(int argc, char *argv[])
 		case O(MANAGE,'f'): /* set faulty */
 			devmode = 'f';
 			continue;
+		case O(INCREMENTAL,'R'):
 		case O(MANAGE,'R'):
 		case O(ASSEMBLE,'R'):
 		case O(BUILD,'R'):
@@ -727,6 +769,7 @@ int main(int argc, char *argv[])
 		case O(MISC,'X'):
 		case O(MISC,'o'):
 		case O(MISC,'w'):
+		case O(MISC,'W'):
 			if (devmode && devmode != opt &&
 			    (devmode == 'E' || (opt == 'E' && devmode != 'Q'))) {
 				fprintf(stderr, Name ": --examine/-E cannot be given with -%c\n",
@@ -776,9 +819,15 @@ int main(int argc, char *argv[])
 			backup_file = optarg;
 			continue;
 
-		case O(GROW,'b'):
 		case O(BUILD,'b'):
 		case O(CREATE,'b'): /* here we create the bitmap */
+			if (strcmp(optarg, "none") == 0) {
+				fprintf(stderr, Name ": '--bitmap none' only"
+					" support for --grow\n");
+				exit(2);
+			}
+			/* FALL THROUGH */
+		case O(GROW,'b'):
 			if (strcmp(optarg, "internal")== 0 ||
 			    strcmp(optarg, "none")== 0 ||
 			    strchr(optarg, '/') != NULL) {
@@ -803,6 +852,7 @@ int main(int argc, char *argv[])
 			bitmap_chunk = bitmap_chunk ? bitmap_chunk * 1024 : 512;
 			continue;
 
+		case O(GROW, WriteBehind):
 		case O(BUILD, WriteBehind):
 		case O(CREATE, WriteBehind): /* write-behind mode */
 			write_behind = DEFAULT_MAX_WRITE_BEHIND;
@@ -815,6 +865,10 @@ int main(int argc, char *argv[])
 				}
 			}
 			continue;
+
+		case O(INCREMENTAL, 'r'):
+			rebuild_map = 1;
+			continue;
 		}
 		/* We have now processed all the valid options. Anything else is
 		 * an error
@@ -843,8 +897,9 @@ int main(int argc, char *argv[])
 			case MISC     : help_text = Help_misc; break;
 			case MONITOR  : help_text = Help_monitor; break;
 			case GROW     : help_text = Help_grow; break;
+			case INCREMENTAL:help_text= Help_incr; break;
 			}
-		fputs(help_text,stderr);
+		fputs(help_text,stdout);
 		exit(0);
 	}
 
@@ -858,6 +913,19 @@ int main(int argc, char *argv[])
 		fputs(Usage, stderr);
 		exit(2);
 	}
+
+	if (symlinks) {
+		struct createinfo *ci = conf_get_create_info();
+
+		if (strcasecmp(symlinks, "yes") == 0)
+			ci->symlinks = 1;
+		else if (strcasecmp(symlinks, "no") == 0)
+			ci->symlinks = 0;
+		else {
+			fprintf(stderr, Name ": option --symlinks must be 'no' or 'yes'\n");
+			exit(2);
+		}
+	}
 	/* Ok, got the option parsing out of the way
 	 * hopefully it's mostly right but there might be some stuff
 	 * missing
@@ -875,7 +943,7 @@ int main(int argc, char *argv[])
 			exit(2);
 		}
 		if ((int)ident.super_minor == -2 && autof) {
-			fprintf(stderr, Name ": --super-minor=dev is incompatible with --auto\n");	
+			fprintf(stderr, Name ": --super-minor=dev is incompatible with --auto\n");
 			exit(2);
 		}
 		if (mode == MANAGE || mode == GROW)
@@ -944,7 +1012,7 @@ int main(int argc, char *argv[])
 					devlist->devname);
 				rv |= 1;
 			} else {
-				mdfd = open_mddev(devlist->devname, 
+				mdfd = open_mddev(devlist->devname,
 						  array_ident->autof ? array_ident->autof : autof);
 				if (mdfd < 0)
 					rv |= 1;
@@ -976,7 +1044,7 @@ int main(int argc, char *argv[])
 					rv |= 1;
 					continue;
 				}
-				mdfd = open_mddev(dv->devname, 
+				mdfd = open_mddev(dv->devname,
 						  array_ident->autof ?array_ident->autof : autof);
 				if (mdfd < 0) {
 					rv |= 1;
@@ -1013,7 +1081,7 @@ int main(int argc, char *argv[])
 				}
 				if (ioctl(mdfd, GET_ARRAY_INFO, &array)>=0)
 					/* already assembled, skip */
-					;
+					cnt++;
 				else {
 					rv |= Assemble(ss, array_list->devname, mdfd,
 						       array_list,
@@ -1084,6 +1152,11 @@ int main(int argc, char *argv[])
 			rv = 1;
 			break;
 		}
+		if (raiddisks == 0) {
+			fprintf(stderr, Name ": no raid-disks specified.\n");
+			rv = 1;
+			break;
+		}
 
 		if (bitmap_file) {
 			if (strcmp(bitmap_file, "internal")==0) {
@@ -1103,14 +1176,19 @@ int main(int argc, char *argv[])
 			rv = 1;
 			break;
 		}
+		if (raiddisks == 0) {
+			fprintf(stderr, Name ": no raid-disks specified.\n");
+			rv = 1;
+			break;
+		}
 
 		rv = Create(ss, devlist->devname, mdfd, chunk, level, layout, size<0 ? 0 : size,
 			    raiddisks, sparedisks, ident.name, homehost,
+			    ident.uuid_set ? ident.uuid : NULL,
 			    devs_found-1, devlist->next, runstop, verbose-quiet, force, assume_clean,
 			    bitmap_file, bitmap_chunk, write_behind, delay);
 		break;
 	case MISC:
-
 		if (devmode == 'E') {
 			if (devlist == NULL && !scan) {
 				fprintf(stderr, Name ": No devices to examine\n");
@@ -1122,7 +1200,11 @@ int main(int argc, char *argv[])
 				fprintf(stderr, Name ": No devices listed in %s\n", configfile?configfile:DefaultConfFile);
 				exit(1);
 			}
-			rv = Examine(devlist, scan?(verbose>1?0:verbose+1):brief, scan, SparcAdjust, ss, homehost);
+			if (brief && verbose)
+				brief = 2;
+			rv = Examine(devlist, scan?(verbose>1?0:verbose+1):brief,
+				     export, scan,
+				     SparcAdjust, ss, homehost);
 		} else {
 			if (devlist == NULL) {
 				if (devmode=='D' && scan) {
@@ -1137,7 +1219,8 @@ int main(int argc, char *argv[])
 								e->dev);
 							continue;
 						}
-						rv |= Detail(name, verbose>1?0:verbose+1, test, homehost);
+						rv |= Detail(name, verbose>1?0:verbose+1,
+							     export, test, homehost);
 						put_md_name(name);
 					}
 				} else	if (devmode == 'S' && scan) {
@@ -1182,13 +1265,18 @@ int main(int argc, char *argv[])
 			for (dv=devlist ; dv; dv=dv->next) {
 				switch(dv->disposition) {
 				case 'D':
-					rv |= Detail(dv->devname, brief?1+verbose:0, test, homehost); continue;
+					rv |= Detail(dv->devname,
+						     brief?1+verbose:0,
+						     export, test, homehost);
+					continue;
 				case 'K': /* Zero superblock */
 					rv |= Kill(dv->devname, force, quiet); continue;
 				case 'Q':
 					rv |= Query(dv->devname); continue;
 				case 'X':
 					rv |= ExamineBitmap(dv->devname, brief, ss); continue;
+				case 'W':
+					rv |= Wait(dv->devname); continue;
 				}
 				mdfd = open_mddev(dv->devname, 1);
 				if (mdfd>=0) {
@@ -1226,7 +1314,7 @@ int main(int argc, char *argv[])
 
 	case GROW:
 		if (devs_found > 1) {
-			
+
 			/* must be '-a'. */
 			if (size >= 0 || raiddisks) {
 				fprintf(stderr, Name ": --size, --raiddisks, and --add are exclusing in --grow mode\n");
@@ -1254,6 +1342,38 @@ int main(int argc, char *argv[])
 		} else
 			fprintf(stderr, Name ": no changes to --grow\n");
 		break;
+	case INCREMENTAL:
+		if (rebuild_map) {
+			RebuildMap();
+		}
+		if (scan) {
+			if (runstop <= 0) {
+				fprintf(stderr, Name
+			 ": --incremental --scan meaningless without --run.\n");
+				break;
+			}
+			rv = IncrementalScan(verbose);
+		}
+		if (!devlist) {
+			if (!rebuild_map && !scan) {
+				fprintf(stderr, Name
+					": --incremental requires a device.\n");
+				rv = 1;
+			}
+			break;
+		}
+		if (devlist->next) {
+			fprintf(stderr, Name
+			       ": --incremental can only handle one device.\n");
+			rv = 1;
+			break;
+		}
+		rv = Incremental(devlist->devname, verbose-quiet, runstop,
+				 ss, homehost, autof);
+		break;
+	case AUTODETECT:
+		autodetect();
+		break;
 	}
 	exit(rv);
 }
